from typing import Dict

from jinad.models.custom import build_pydantic_model

PodModel = build_pydantic_model(model_name='PodModel',
                                module='pod')


class RemotePodModel:
    args: Dict = {
        'head': None,
        'tail': None,
        'peas': [PodModel]
    }



from collections import namedtuple

from fastapi import FastAPI

from jinad.api.endpoints import common_router, flow, pod, pea, logs
from jinad.config import jinad_config, fastapi_config, server_config, openapitags_config


def get_app():
    context = namedtuple('context', ['router', 'openapi_tags', 'tags'])
    _all_routers = {
        'flow': context(router=flow.router,
                        openapi_tags=openapitags_config.FLOW_API_TAGS,
                        tags=[openapitags_config.FLOW_API_TAGS[0]['name']]),
        'pod': context(router=pod.router,
                       openapi_tags=openapitags_config.POD_API_TAGS,
                       tags=[openapitags_config.POD_API_TAGS[0]['name']]),
        'pea': context(router=pea.router,
                       openapi_tags=openapitags_config.PEA_API_TAGS,
                       tags=[openapitags_config.PEA_API_TAGS[0]['name']])
    }
    app = FastAPI(
        title=fastapi_config.NAME,
        description=fastapi_config.DESCRIPTION,
        version=fastapi_config.VERSION
    )
    app.include_router(router=common_router,
                       prefix=fastapi_config.PREFIX)
    app.include_router(router=logs.router,
                       prefix=fastapi_config.PREFIX)
    if jinad_config.CONTEXT == 'all':
        for _current_router in _all_routers.values():
            app.include_router(router=_current_router.router,
                               tags=_current_router.tags,
                               prefix=fastapi_config.PREFIX)
    else:
        _current_router = _all_routers[jinad_config.CONTEXT]
        app.openapi_tags = _current_router.openapi_tags
        app.include_router(router=_current_router.router,
                           tags=_current_router.tags,
                           prefix=fastapi_config.PREFIX)
    return app


def write_openapi_schema(filename='schema.json'):
    import json
    app = get_app()
    schema = app.openapi()
    with open(filename, 'w') as f:
        json.dump(schema, f)


def uvicorn_serve(app: 'FastAPI'):
    from uvicorn import Config, Server
    config = Config(app=app,
                    host=server_config.HOST,
                    port=server_config.PORT,
                    loop='uvloop',
                    log_level='error')
    server = Server(config=config)
    server.run()


def start():
    app = get_app()
    uvicorn_serve(app=app)


if __name__ == "__main__":
    start()



from jina import __version__ as jina_version
from jina.logging import JinaLogger
from fastapi import status, APIRouter

from jinad.config import server_config

logger = JinaLogger(context='👻 JINAD')
common_router = APIRouter()


@common_router.on_event('startup')
async def startup():
    logger.success(f'Uvicorn + FastAPI running on {server_config.HOST}:{server_config.PORT}')
    logger.success('Welcome to Jina daemon - the remote manager for jina!')


@common_router.get(
    path='/alive',
    summary='Get status of jinad',
    status_code=status.HTTP_200_OK
)
async def _status():
    """
    Used to check if the api is running (returns 200 & jina version)
    """
    # TODO(Deepankar): should we add versions of executors?
    return {
        'status_code': status.HTTP_200_OK,
        'jina_version': jina_version
    }



import argparse
import os
from typing import Dict

from jina.helper import get_random_identity
from fastapi import UploadFile

from jinad.models.pea import PeaModel
from jinad.models.pod import PodModel


def get_enum_defaults(parser: argparse.ArgumentParser):
    """ Helper function to get all args that have Enum default values """
    from enum import Enum
    all_args = parser.parse_args([])
    enum_args = {}
    for key in vars(all_args):
        if isinstance(parser.get_default(key), Enum):
            enum_args[key] = parser.get_default(key)
    return enum_args


def handle_enums(args: Dict, parser: argparse.ArgumentParser) -> Dict:
    """ Since REST relies on json, reverse conversion of integers to enums is needed """
    default_enums = get_enum_defaults(parser=parser)
    _args = args.copy()
    if 'log_config' in _args:
        _args['log_config'] = parser.get_default('--log-config')

    for key, value in args.items():
        if key in default_enums:
            _enum_type = type(default_enums[key])
            if isinstance(value, int):
                _args[key] = _enum_type(value)
            elif isinstance(value, str):
                _args[key] = _enum_type.from_string(value)
    return _args


def handle_log_id(args: Dict):
    args['log_id'] = args['identity'] if 'identity' in args else get_random_identity()


def flowpod_to_namespace(args: Dict):
    # TODO: combine all 3 to_namespace methods
    from jina.parser import set_pod_parser
    parser = set_pod_parser()
    pod_args = {}

    for pea_type, pea_args in args.items():
        # this is for pea_type: head & tail when None
        if pea_args is None:
            pod_args[pea_type] = None

        # this is for pea_type: head & tail when not None
        if isinstance(pea_args, dict):
            pea_args = handle_enums(args=pea_args,
                                    parser=parser)
            handle_log_id(args=pea_args)
            pod_args[pea_type] = argparse.Namespace(**pea_args)

        # this is for pea_type: peas (multiple entries)
        if isinstance(pea_args, list):
            pod_args[pea_type] = []
            for pea_arg in pea_args:
                pea_arg = handle_enums(args=pea_arg,
                                       parser=parser)
                handle_log_id(args=pea_arg)
                pod_args[pea_type].append(argparse.Namespace(**pea_arg))

    return pod_args


def basepod_to_namespace(args: PodModel):
    from jina.parser import set_pod_parser
    parser = set_pod_parser()

    if isinstance(args, PodModel):
        pod_args = handle_enums(args=args.dict(),
                                parser=parser)
        handle_log_id(args=pod_args)
        return argparse.Namespace(**pod_args)


def basepea_to_namespace(args: PeaModel):
    from jina.parser import set_pea_parser
    parser = set_pea_parser()

    if isinstance(args, PeaModel):
        pea_args = handle_enums(args=args.dict(),
                                parser=parser)
        handle_log_id(args=pea_args)
        return argparse.Namespace(**pea_args)


def create_meta_files_from_upload(current_file: UploadFile):
    with open(current_file.filename, 'wb') as f:
        f.write(current_file.file.read())


def delete_meta_files_from_upload(current_file: UploadFile):
    if os.path.isfile(current_file.filename):
        os.remove(current_file.filename)



import requests
import argparse
from typing import Union
from pydantic import create_model, validator, Field

JINA_API_URL = 'https://api.jina.ai/latest'


def get_latest_api():
    """Fetches the latest jina cli args"""
    response = requests.get(JINA_API_URL)
    all_cli_args = response.json()
    return all_cli_args


def get_module_args(all_args: list, module: str):
    """Fetches the cli args for modules like `flow`, `pod`"""
    for current_module in all_args['methods']:
        if current_module['name'] == module:
            module_args = current_module
            return module_args


def generate_validator(field: str, choices: list):
    """ Pydantic validator classmethod generator to validate fields exist in choices """
    def validate_arg_choices(v, values):
        if v not in choices:
            raise ValueError(f'Invalid value {v} for field {field}'
                             f'Valid choices are {choices}')
        return v

    validate_arg_choices.__qualname__ = 'validate_' + field
    return validator(field, allow_reuse=True)(validate_arg_choices)


def get_pydantic_fields(config: Union[dict, argparse.ArgumentParser]):
    all_options = {}
    choices_validators = {}

    if isinstance(config, dict):
        for arg in config['options']:
            arg_key = arg['name']
            arg_type = arg['type']
            if arg['choices']:
                choices_validators[f'validator_for_{arg_key}'] = generate_validator(field=arg_key,
                                                                                    choices=arg['choices'])
            if arg_type == 'method':
                arg_type = type(arg['default']) if arg['default'] else int
            arg_type = 'str' if arg_type == 'FileType' else arg_type

            current_field = Field(default=arg['default'],
                                  example=arg['default'],
                                  description=arg['help'])
            all_options[arg_key] = (arg_type, current_field)

    if isinstance(config, argparse.ArgumentParser):
        # Ignoring first 3 as they're generic args
        from jina.parser import KVAppendAction
        for arg in config._actions[3:]:
            arg_key = arg.dest
            arg_type = arg.type
            if arg.choices:
                choices_validators[f'validator_for_{arg_key}'] = generate_validator(field=arg_key,
                                                                                    choices=arg.choices)
            # This is to handle the Enum args (to check if it is a bound method)
            if hasattr(arg_type, '__self__'):
                arg_type = type(arg.default) if arg.default else int
            arg_type = str if isinstance(arg_type, argparse.FileType) else arg_type
            arg_type = dict if type(arg) == KVAppendAction else arg_type

            current_field = Field(default=arg.default,
                                  example=arg.default,
                                  description=arg.help)
            all_options[arg_key] = (arg_type, current_field)

    return all_options, choices_validators


class PydanticConfig:
    arbitrary_types_allowed = True


def build_pydantic_model(kind: str = 'local',
                         model_name: str = 'CustomModel',
                         module: str = 'pod'):
    if kind == 'api':
        all_cli_args = get_latest_api()
        module_args = get_module_args(all_args=all_cli_args,
                                      module=module)
        all_fields, field_validators = get_pydantic_fields(config=module_args)

    elif kind == 'local':
        from jina.parser import set_pea_parser, set_pod_parser, set_flow_parser
        if module == 'pod':
            parser = set_pod_parser()
        elif module == 'pea':
            parser = set_pea_parser()
        elif module == 'flow':
            parser = set_flow_parser()
        all_fields, field_validators = get_pydantic_fields(config=parser)

    return create_model(model_name,
                        **all_fields,
                        __config__=PydanticConfig,
                        __validators__=field_validators)



import uuid
from typing import Dict, List

from fastapi import status, APIRouter, File, UploadFile
from jina.logging import JinaLogger

from jinad.store import pod_store
from jinad.models.pod import PodModel
from jinad.excepts import HTTPException, PodStartException
from jinad.helper import flowpod_to_namespace, basepod_to_namespace, create_meta_files_from_upload

logger = JinaLogger(context='👻 PODAPI')
router = APIRouter()


@router.put(
    path='/upload',
    summary='Upload pod context yamls & pymodules',
)
async def _upload(
    uses_files: List[UploadFile] = File(()),
    pymodules_files: List[UploadFile] = File(())
):
    """

    """
    upload_status = 'nothing to upload'
    if uses_files:
        [create_meta_files_from_upload(current_file) for current_file in uses_files]
        upload_status = 'uploaded'

    if pymodules_files:
        [create_meta_files_from_upload(current_file) for current_file in pymodules_files]
        upload_status = 'uploaded'

    return {
        'status_code': status.HTTP_200_OK,
        'status': upload_status
    }


@router.put(
    path='/pod/cli',
    summary='Create an independent MutablePod',
)
async def _create_independent(
    pod_arguments: PodModel
):
    """
    This is used to create an independent remote MutablePod which stays out of Flow context
    """
    pod_arguments = basepod_to_namespace(args=pod_arguments)

    with pod_store._session():
        try:
            pod_id = pod_store._create(pod_arguments=pod_arguments)
        except PodStartException as e:
            raise HTTPException(status_code=404,
                                detail=f'Pod couldn\'t get started:  {repr(e)}')
        except Exception as e:
            logger.error(f'Got an error while creating a pod {repr(e)}')
            raise HTTPException(status_code=404,
                                detail=f'Something went wrong')
    return {
        'status_code': status.HTTP_200_OK,
        'pod_id': pod_id,
        'status': 'started'
    }


@router.put(
    path='/pod/flow',
    summary='Create a MutablePod via Flow',
)
async def _create_via_flow(
    pod_arguments: Dict
):
    """
    This is used to create a remote MutablePod which gets triggered by a Flow context

    > Shouldn't be created with manual trigger

    Args: pod_arguments (Dict)

        {
            'head': PodModel,
            'tail': PodModel,
            'peas': [PodModel]
        }


    """
    pod_arguments = flowpod_to_namespace(args=pod_arguments)

    with pod_store._session():
        try:
            pod_id = pod_store._create(pod_arguments=pod_arguments)
        except PodStartException as e:
            raise HTTPException(status_code=404,
                                detail=f'Pod couldn\'t get started:  {repr(e)}')
        except Exception as e:
            logger.error(f'Got an error while creating a pod {repr(e)}')
            raise HTTPException(status_code=404,
                                detail=f'Something went wrong')
    return {
        'status_code': status.HTTP_200_OK,
        'pod_id': pod_id,
        'status': 'started'
    }


@router.delete(
    path='/pod',
    summary='Delete pod',
)
async def _delete(
    pod_id: uuid.UUID
):
    """Close Pod context
    """
    with pod_store._session():
        try:
            pod_store._delete(pod_id=pod_id)
            return {
                'status_code': status.HTTP_200_OK
            }
        except KeyError:
            raise HTTPException(status_code=status.HTTP_404_NOT_FOUND,
                                detail=f'Pod ID {pod_id} not found! Please create a new Pod')


@router.on_event('shutdown')
def _shutdown():
    with pod_store._session():
        pod_store._delete_all()



import uuid
from contextlib import contextmanager
from tempfile import SpooledTemporaryFile
from typing import List, Dict, Union
from argparse import Namespace

from fastapi import UploadFile
from jina.flow import Flow
from jina.jaml import JAML
from jina.helper import colored, get_random_identity
from jina.logging import JinaLogger
from jina.peapods import Runtime, Pod

from jinad.models.pod import PodModel
from jinad.helper import create_meta_files_from_upload, delete_meta_files_from_upload
from jinad.excepts import PeaFailToStart, FlowYamlParseException, FlowCreationException, \
    FlowStartException, PodStartException, PeaStartException, FlowBadInputException


class InMemoryStore:
    _store = {}
    # TODO: Implement fastapi based oauth/bearer security here
    credentials = 'foo:bar'
    _session_token = None
    logger = JinaLogger(context='🏪 STORE')

    @contextmanager
    def _session(self):
        if self._session_token:
            yield
            return

        self._session_token = self._login(self.credentials)
        try:
            yield
        finally:
            self._logout(self._session_token)

    # TODO: implement login-logout here to manage session token
    def _login(self, creds):
        token = hash(creds)
        self.logger.debug(f'LOGIN: {token}')
        return token

    def _logout(self, token):
        self.logger.debug(f'LOGOUT: {token}')

    def _create(self, **kwargs):
        raise NotImplementedError

    def _start(self, context):
        return context.start()

    def _close(self, context):
        context.close()

    def _delete(self, id: uuid.UUID):
        raise NotImplementedError

    def _delete_all(self):
        for _id in self._store.copy().keys():
            self._delete(_id)


class InMemoryFlowStore(InMemoryStore):

    def _create(self,
                config: Union[str, SpooledTemporaryFile, List[PodModel]] = None,
                files: List[UploadFile] = None):
        """ Creates Flow using List[PodModel] or yaml spec """
        # This makes sure `uses` & `py_modules` are created locally in `cwd`
        # TODO: Handle file creation, deletion better
        if files:
            [create_meta_files_from_upload(current_file) for current_file in files]

        # FastAPI treats UploadFile as a tempfile.SpooledTemporaryFile
        # I think this needs to be handled by some `FlowBuilder` class, transfrom a yaml_load to a config
        if isinstance(config, str) or isinstance(config, SpooledTemporaryFile):
            yamlspec = config.read().decode() if isinstance(config, SpooledTemporaryFile) else config
            try:
                JAML.register(Flow)
                flow = JAML.load(yamlspec)
            except Exception as e:
                self.logger.error(f'Got error while loading from yaml {repr(e)}')
                raise FlowYamlParseException
        elif isinstance(config, list):
            try:
                flow = Flow()
                # it is strange to build from a given flow, it seems like a lazy construction pattern could be used?
                flow = self._build_with_pods(flow=flow,
                                             pod_args=config)
            except Exception as e:
                self.logger.error(f'Got error while creating flows via pods: {repr(e)}')
                raise FlowCreationException
        else:
            raise FlowBadInputException(f'Not valid Flow config input {type(config)}')

        try:
            flow.args.log_id = flow.args.identity if 'identity' in flow.args else get_random_identity()
            flow_id = uuid.UUID(flow.args.log_id)
            flow = self._start(context=flow)
        except PeaFailToStart as e:
            self.logger.critical(f'Flow couldn\'t get started - Invalid Pod {repr(e)} ')
            self.logger.critical('Possible causes - invalid/not uploaded pod yamls & pymodules')
            # TODO: Send correct error message
            raise FlowStartException(repr(e))
        except Exception as e:
            self.logger.critical(f'Got following error while starting the flow: {repr(e)}')
            raise FlowStartException(repr(e))

        self._store[flow_id] = {}
        self._store[flow_id]['flow'] = flow
        self._store[flow_id]['files'] = files
        self.logger.info(f'Started flow with flow_id {colored(flow_id, "cyan")}')
        return flow_id, flow.host, flow.port_expose

    def _build_with_pods(self,
                         flow: Flow,
                         pod_args: List[PodModel]):
        """ Since we rely on PodModel, this can accept all params that a Pod can accept """
        for current_pod_args in pod_args:
            _current_pod_args = current_pod_args.dict()
            _current_pod_args.pop('log_config')
            flow = flow.add(**_current_pod_args)
        return flow

    def _get(self,
             flow_id: uuid.UUID,
             yaml_only: bool = False):
        """ Fetches a Flow from the store """
        if flow_id not in self._store:
            raise KeyError(f'{flow_id} not found')

        if 'flow' in self._store[flow_id]:
            flow = self._store[flow_id]['flow']

        return flow.host, flow.port_expose, flow.yaml_spec

    def _delete(self, flow_id: uuid.UUID):
        """ Closes a Flow context & deletes from store """
        if flow_id not in self._store:
            raise KeyError(f'flow_id {flow_id} not found in store. please create one!')
        flow = self._store.pop(flow_id)

        if 'flow' in flow:
            self._close(context=flow['flow'])

        if 'files' in flow and flow['files']:
            for current_file in flow['files']:
                delete_meta_files_from_upload(current_file=current_file)

        self.logger.info(f'Closed flow with flow_id {colored(flow_id, "cyan")}')


class InMemoryPodStore(InMemoryStore):

    def _create(self, pod_arguments: Union[Dict, Namespace]):
        """ Creates a Pod via Flow or via CLI """

        try:
            pod_id = uuid.UUID(pod_arguments.log_id) if isinstance(pod_arguments, Namespace) \
                else uuid.UUID(pod_arguments['peas'][0].log_id)

            pod = Pod(args=pod_arguments, allow_remote=False)
            pod = self._start(context=pod)
        except Exception as e:
            self.logger.critical(f'Got following error while starting the pod: {repr(e)}')
            raise PodStartException(repr(e))

        self._store[pod_id] = {}
        self._store[pod_id]['pod'] = pod
        self.logger.info(f'Started pod with pod_id {colored(pod_id, "cyan")}')
        return pod_id

    def _delete(self, pod_id: uuid.UUID):
        """ Closes a Pod context & deletes from store """
        if pod_id not in self._store:
            raise KeyError(f'pod_id {pod_id} not found in store. please create one!')
        pod = self._store.pop(pod_id)

        if 'pod' in pod:
            self._close(context=pod['pod'])

        if 'files' in pod:
            for current_file in pod['files']:
                delete_meta_files_from_upload(current_file=current_file)

        self.logger.info(f'Closed pod with pod_id {colored(pod_id, "cyan")}')


class InMemoryPeaStore(InMemoryStore):
    """ Creates Pea on remote  """

    # TODO: Merge this with InMemoryPodStore
    def _create(self, pea_arguments: Union[Dict, Namespace]):
        try:
            pea_id = uuid.UUID(pea_arguments.log_id) if isinstance(pea_arguments, Namespace) \
                else uuid.UUID(pea_arguments['log_id'])
            pea = Runtime(args=pea_arguments, allow_remote=False)
            pea = self._start(context=pea)
        except Exception as e:
            self.logger.critical(f'Got following error while starting the pea: {repr(e)}')
            raise PeaStartException(repr(e))

        self._store[pea_id] = {}
        self._store[pea_id]['pea'] = pea
        self.logger.info(f'Started pea with pea_id {colored(pea_id, "cyan")}')
        return pea_id

    def _delete(self, pea_id: uuid.UUID):
        """ Closes a Pea context & deletes from store """
        if pea_id not in self._store:
            raise KeyError(f'pea_id {pea_id} not found in store. please create one!')
        pea = self._store.pop(pea_id)

        if 'pea' in pea:
            self._close(context=pea['pea'])

        if 'files' in pea:
            for current_file in pea['files']:
                delete_meta_files_from_upload(current_file=current_file)

        self.logger.info(f'Closed pea with pea_id {colored(pea_id, "cyan")}')


flow_store = InMemoryFlowStore()
pod_store = InMemoryPodStore()
pea_store = InMemoryPeaStore()



import uuid
import time
import json
import asyncio
from pathlib import Path
from typing import Optional

from jina.logging import JinaLogger
from fastapi import APIRouter, WebSocket, WebSocketDisconnect

from jinad.config import log_config
from jinad.excepts import HTTPException, TimeoutException, ClientExit


logger = JinaLogger(context='👻 LOGS')
router = APIRouter()

async def tail(file_handler, line_num_from=0, timeout=5):
    """ asynchronous tail file """
    line_number = 0
    last_log_time = time.time()
    while time.time() - last_log_time < timeout:
        for line in file_handler:
            line_number += 1
            if line_number < line_num_from:
                continue
            yield line_number, line
            last_log_time = time.time()
            await asyncio.sleep(0.01)
    else:
        raise TimeoutException()


@router.websocket(
    path='/wslog/{log_id}'
)
async def _websocket_logs(
    websocket: WebSocket,
    log_id: uuid.UUID,
    timeout: Optional[int] = 5,
    exit_text: Optional[str] = 'exit'
):
    """
    # TODO: Swagger doesn't support websocket based docs
    Websocket endpoint to stream logs from fluentd file

    ```
    `log_id`: uuid of the flow/pod/pea
    `timeout`: max time difference b/w successive log lines (helps in client disconnection)
    `exit_text`: exit the connection if this text is found in the message
    ```
    """
    file_path = log_config.PATH % log_id
    if not Path(file_path).is_file():
        raise HTTPException(status_code=404,
                            detail=f'No logs found')

    await websocket.accept()
    client_host = websocket.client.host
    client_port = websocket.client.port
    logger.info(f'Client {client_host}:{client_port} got connected!')
    data = await websocket.receive_json()
    logger.debug(f'received the first message: {data}')
    line_num_from = int(data.get('from', 0))
    logs_to_be_sent = {}

    try:
        with open(file_path) as fp:
            async for line_number, line in tail(file_handler=fp, timeout=timeout):
                if line_number > line_num_from:
                    logs_to_be_sent[line_number] = line
                    logger.info(f'Sending logs {logs_to_be_sent}')
                    await websocket.send_text(json.dumps(logs_to_be_sent))
                    data = await websocket.receive_json()
                    if exit_text in data:
                        raise ClientExit
                    logs_to_be_sent = {}
    except WebSocketDisconnect:
        logger.info(f'Client {client_host}:{client_port} got disconnected!')
    except TimeoutException:
        await websocket.close()
        logger.info(f'No logs found in last {timeout} secs. Timing out!')
        logger.info(f'Closing client {client_host}:{client_port}!')
    except ClientExit:
        await websocket.close()
        logger.info('Client asked to exit!')
        logger.info(f'Closing client {client_host}:{client_port}!')



import json
import uuid
from typing import List, Union

from fastapi import status, APIRouter, Body, Response, File, UploadFile
from jina.parser import set_client_cli_parser
from jina.helper import ArgNamespace
from jina.logging import JinaLogger
from jina.clients import Client

from jinad.store import flow_store
from jinad.models.pod import PodModel
from jinad.excepts import FlowYamlParseException, FlowCreationException, FlowStartException, \
    HTTPException

logger = JinaLogger(context='👻 FLOWAPI')
router = APIRouter()


@router.put(
    path='/flow/pods',
    summary='Build & start a Flow using Pods',
)
async def _create_from_pods(
    pods: Union[List[PodModel]] = Body(...,
                                       example=json.loads(PodModel().json()))
):
    """
    Build a Flow using a list of `PodModel`

        [
            {
                "name": "pod1",
                "uses": "_pass"
            },
            {
                "name": "pod2",
                "uses": "_pass",
                "host": "10.18.3.127",
                "port_expose": 8000
            }
        ]
    """
    with flow_store._session():
        try:
            flow_id, host, port_expose = flow_store._create(config=pods)
        except FlowCreationException:
            raise HTTPException(status_code=404,
                                detail=f'Bad pods args')
        except FlowStartException:
            raise HTTPException(status_code=404,
                                detail=f'Flow couldn\'t get started')
    return {
        'status_code': status.HTTP_200_OK,
        'flow_id': flow_id,
        'host': host,
        'port': port_expose,
        'status': 'started'
    }


@router.put(
    path='/flow/yaml',
    summary='Build & start a Flow using YAML',
)
async def _create_from_yaml(
    yamlspec: UploadFile = File(...),
    uses_files: List[UploadFile] = File(()),
    pymodules_files: List[UploadFile] = File(())
):
    """
    Build a flow using [Flow YAML](https://docs.jina.ai/chapters/yaml/yaml.html#flow-yaml-sytanx)

    > Upload Flow yamlspec (`yamlspec`)

    > Yamls that Pods use (`uses_files`) (Optional)

    > Python modules (`pymodules_files`) that the Pods use (Optional)

    **yamlspec**:

        !Flow
        with:
            rest_api: true
            compress_hwm: 1024
        pods:
            encode:
                uses: helloworld.encoder.yml
                parallel: 2
            index:
                uses: helloworld.indexer.yml
                shards: 2
                separated_workspace: true

    **uses_files**: `helloworld.encoder.yml`

        !MyEncoder
        metas:
            name: myenc
            workspace: /tmp/blah
            py_modules: components.py
        requests:
            on:
                [IndexRequest, SearchRequest]:
                - !Blob2PngURI {}
                - !EncodeDriver {}
                - !ExcludeQL
                with:
                    fields:
                        - buffer
                        - chunks

    **uses_files**: `helloworld.indexer.yml`

        !CompoundIndexer
        components:
        - !NumpyIndexer
            with:
                index_filename: vec.gz
            metas:
                name: vecidx
                workspace: /tmp/blah
        - !BinaryPbIndexer
            with:
                index_filename: chunk.gz
            metas:
                name: chunkidx
                workspace: /tmp/blah
        metas:
            name: chunk_indexer
            workspace: /tmp/blah

    **pymodules_files**: `components.py`

        class MyEncoder(BaseImageEncoder):
            def __init__(self):
                ...

    """

    with flow_store._session():
        try:
            flow_id, host, port_expose = flow_store._create(config=yamlspec.file,
                                                            files=list(uses_files) + list(pymodules_files))
        except FlowYamlParseException:
            raise HTTPException(status_code=404,
                                detail=f'Invalid yaml file.')
        except FlowStartException as e:
            raise HTTPException(status_code=404,
                                detail=f'Flow couldn\'t get started:  {repr(e)}')

    return {
        'status_code': status.HTTP_200_OK,
        'flow_id': flow_id,
        'host': host,
        'port': port_expose,
        'status': 'started'
    }


@router.get(
    path='/flow/{flow_id}',
    summary='Get Flow information',
)
async def _fetch(
    flow_id: uuid.UUID,
    yaml_only: bool = False
):
    """
    Get Flow information using `flow_id`.

    Following details are sent:
    - Flow YAML
    - Gateway host
    - Gateway port
    """
    try:
        with flow_store._session():
            host, port_expose, yaml_spec = flow_store._get(flow_id=flow_id)

        if yaml_only:
            return Response(content=yaml_spec,
                            media_type='application/yaml')

        return {
            'status_code': status.HTTP_200_OK,
            'yaml': yaml_spec,
            'host': host,
            'port': port_expose
        }
    except KeyError:
        raise HTTPException(status_code=status.HTTP_404_NOT_FOUND,
                            detail=f'Flow ID {flow_id} not found! Please create a new Flow')


@router.get(
    path='/ping',
    summary='Connect to Flow gateway',
)
async def _ping(
    host: str,
    port: int
):
    """
    Ping to check if we can connect to gateway via gRPC `host:port`

    Note: Make sure Flow is running
    """
    kwargs = {'port_expose': port, 'host': host}
    _, args, _ = ArgNamespace.get_parsed_args(kwargs, set_client_cli_parser())
    client = Client(args)
    try:
        client.index(input_fn=['abc'])
        return {
            'status_code': status.HTTP_200_OK,
            'detail': 'connected'
        }
    except Exception:
        raise HTTPException(status_code=status.HTTP_404_NOT_FOUND,
                            detail=f'Cannot connect to GRPC Server on {host}:{port}')


@router.delete(
    path='/flow',
    summary='Close Flow context',
)
async def _delete(
    flow_id: uuid.UUID
):
    """
    Close Flow context
    """
    with flow_store._session():
        try:
            flow_store._delete(flow_id=flow_id)
            return {
                'status_code': status.HTTP_200_OK
            }
        except KeyError:
            raise HTTPException(status_code=status.HTTP_404_NOT_FOUND,
                                detail=f'Flow ID {flow_id} not found! Please create a new Flow')


@router.on_event('shutdown')
def _shutdown():
    with flow_store._session():
        flow_store._delete_all()



from jina.excepts import PeaFailToStart, GRPCServerError
from fastapi.exceptions import HTTPException


class FlowYamlParseException(Exception):
    """ Exception during loading yaml file for Flow creation"""


class FlowCreationException(Exception):
    """ Exception during flow creation via pods"""


class FlowBadInputException(Exception):
    """ Exception during loading Flow, no valid configuration"""


class FlowStartException(Exception):
    """ Exception during flow start"""


class PodStartException(Exception):
    """ Exception during pod start """


class PeaStartException(Exception):
    """ Exception during pod start """


class TimeoutException(Exception):
    """ Exception to raise for successive log line timeout """


class ClientExit(Exception):
    """ Exception for websocket client closure """



import os
from setuptools import setup
from setuptools import find_packages

try:
    pkg_name = 'jinad'
    libinfo_py = os.path.join(pkg_name, '__init__.py')
    libinfo_content = open(libinfo_py, 'r', encoding='utf8').readlines()
    version_line = [l.strip() for l in libinfo_content if l.startswith('__version__')][0]
    exec(version_line)  # gives __version__
except FileNotFoundError:
    __version__ = '0.0.0'

try:
    with open('README.md', encoding='utf8') as fp:
        _long_description = fp.read()
except FileNotFoundError:
    _long_description = ''

install_requires = ['fastapi', 'uvicorn', 'pydantic', 'python-multipart', 'requests']
extras_require = {'all': ['flaky', 'pytest', 'pytest-asyncio', 'pytest-cov']}
jinaver = os.environ.get('JINAVER', 'jina')
if jinaver:
    install_requires.append(jinaver)

setup(
    name=pkg_name,
    version=__version__,
    description='Jinad is the daemon tool for running Jina on remote machines. Jina is the cloud-native neural search solution powered by the state-of-the-art AI and deep learning',
    packages=find_packages(exclude=("tests*",)),
    author='Jina Dev Team',
    author_email='dev-team@jina.ai',
    license='Apache 2.0',
    url='https://opensource.jina.ai',
    download_url='https://github.com/jina-ai/jinad/tags',
    python_requires='>=3.7',
    long_description=_long_description,
    long_description_content_type='text/markdown',
    setup_requires=[
        'setuptools>=18.0',
    ],
    entry_points={
        'console_scripts': ['jinad=jinad.main:start'],
    },
    install_requires=install_requires,
    extras_require=extras_require,
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'Intended Audience :: Education',
        'Intended Audience :: Science/Research',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Unix Shell',
        'Environment :: Console',
        'License :: OSI Approved :: Apache Software License',
        'Operating System :: OS Independent',
        'Topic :: Database :: Database Engines/Servers',
        'Topic :: Scientific/Engineering :: Artificial Intelligence',
        'Topic :: Internet :: WWW/HTTP :: Indexing/Search',
        'Topic :: Scientific/Engineering :: Image Recognition',
        'Topic :: Multimedia :: Video',
        'Topic :: Scientific/Engineering',
        'Topic :: Scientific/Engineering :: Mathematics',
        'Topic :: Software Development',
        'Topic :: Software Development :: Libraries',
        'Topic :: Software Development :: Libraries :: Python Modules',
    ],
    keywords='jina cloud-native neural-search query search index elastic neural-network encoding '
             'embedding serving docker container image video audio deep-learning',
)



from . import __version__, __prefix__
from pydantic import BaseSettings, validator


class BaseConfig(BaseSettings):
    class Config:
        env_prefix = 'JINAD_'


class FastAPIConfig(BaseConfig):
    NAME: str = 'Jina Remote Manager'
    DESCRIPTION: str = 'REST API for managing Jina on Remote'
    VERSION: str = __version__
    PREFIX: str = '/' + __prefix__


class OpenAPITags(BaseConfig):
    API_TAGS: list = [{
        "name": "Jina Remote Management",
        "description": "API to invoke remote Flows/Pods/Peas",
        "externalDocs": {
            "description": "Jina Remote Context Manager",
            "url": "https://docs.jina.ai/",
        },
    }]
    FLOW_API_TAGS: list = [{
        "name": "Remote Flow Manager",
        "description": "API to invoke local/remote Flows",
        "externalDocs": {
            "description": "Jina Flow Context Manager",
            "url": "https://docs.jina.ai/chapters/flow/index.html",
        },
    }]
    POD_API_TAGS: list = [{
        "name": "Remote Pod Manager",
        "description": "API to invoke remote Pods (__should be used by Flow APIs only__)",
        "externalDocs": {
            "description": "Jina 101",
            "url": "https://docs.jina.ai/chapters/101/.sphinx.html",
        },
    }]
    PEA_API_TAGS: list = [{
        "name": "Remote Pea Manager",
        "description": "API to invoke remote Peas",
        "externalDocs": {
            "description": "Jina 101",
            "url": "https://docs.jina.ai/chapters/101/.sphinx.html",
        },
    }]
    LOG_API_TAGS: list = [{
        "name": "logs",
        "description": "Endpoint to get streaming logs from flows/pods",
    }]


class ServerConfig(BaseConfig):
    # TODO: check if HOST can be a ipaddress.IPv4Address
    HOST: str = '0.0.0.0'
    PORT: int = 8000


class JinaDConfig(BaseConfig):
    CONTEXT: str = 'all'

    @validator('CONTEXT')
    def validate_name(cls, value):
        if value.lower() not in ['all', 'flow', 'pod', 'pea']:
            raise ValueError('CONTEXT must be either all, flow or pod or pea')
        return value.lower()


class LogConfig(BaseConfig):
    # TODO: Read config from some file
    PATH: str = '/tmp/jina-log/%s/log.log'


jinad_config = JinaDConfig()
log_config = LogConfig()
fastapi_config = FastAPIConfig()
server_config = ServerConfig()
openapitags_config = OpenAPITags()
